# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# Package module:   DOM: IIR to *** translations.
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
"""
This module offers helper functions to translate often used IIR substructures to pyGHDL.dom (pyVHDLModel) constructs.
"""

from typing import List, Generator, Type, Dict

from pyTooling.Decorators import export
from pyTooling.Warning import WarningCollector

from pyVHDLModel.Base import ModelEntity, Direction, ExpressionUnion
from pyVHDLModel.Name import Name
from pyVHDLModel.Symbol import Symbol
from pyVHDLModel.Association import AssociationItem
from pyVHDLModel.Interface import GenericInterfaceItemMixin, PortInterfaceItemMixin, ParameterInterfaceItemMixin
from pyVHDLModel.Type import BaseType
from pyVHDLModel.Sequential import SequentialStatement
from pyVHDLModel.Concurrent import ConcurrentStatement

from pyGHDL.libghdl import utils, name_table
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes
from pyGHDL.dom.Sequential import (
    IfStatement,
    ForLoopStatement,
    CaseStatement,
    SequentialReportStatement,
    SequentialAssertStatement,
    WaitStatement,
    SequentialSimpleSignalAssignment,
    NullStatement,
    ExitStatement,
    SequentialProcedureCall,
)

from pyGHDL.dom import Position, DOMException
from pyGHDL.dom._Utils import GetNameOfNode, GetIirKindOfNode
from pyGHDL.dom.Name import (
    SimpleName,
    SelectedName,
    AttributeName,
    ParenthesisName,
    AllName,
    OpenName,
)
from pyGHDL.dom.Symbol import (
    SimpleObjectOrFunctionCallSymbol,
    SimpleSubtypeSymbol,
    ConstrainedArraySubtypeSymbol,
    IndexedObjectOrFunctionCallSymbol,
    ConstrainedScalarSubtypeSymbol,
    ConstrainedRecordSubtypeSymbol,
    RecordElementSymbol,
)
from pyGHDL.dom.Type import (
    IntegerType,
    Subtype,
    ArrayType,
    RecordType,
    EnumeratedType,
    AccessType,
    ProtectedType,
    ProtectedTypeBody,
    FileType,
    PhysicalType,
    IncompleteType,
)
from pyGHDL.dom.Range import Range
from pyGHDL.dom.Literal import (
    IntegerLiteral,
    CharacterLiteral,
    FloatingPointLiteral,
    StringLiteral,
    PhysicalIntegerLiteral,
    PhysicalFloatingLiteral,
    NullLiteral,
)
from pyGHDL.dom.Object import Variable
from pyGHDL.dom.Expression import (
    SubtractionExpression,
    AdditionExpression,
    MultiplyExpression,
    DivisionExpression,
    InverseExpression,
    ExponentiationExpression,
    Aggregate,
    NegationExpression,
    ParenthesisExpression,
    ConcatenationExpression,
    QualifiedExpression,
    ModuloExpression,
    RemainderExpression,
    AndExpression,
    NandExpression,
    OrExpression,
    NorExpression,
    XorExpression,
    XnorExpression,
    EqualExpression,
    UnequalExpression,
    LessThanExpression,
    GreaterThanExpression,
    GreaterEqualExpression,
    LessEqualExpression,
    ShiftLeftLogicExpression,
    ShiftRightLogicExpression,
    ShiftLeftArithmeticExpression,
    ShiftRightArithmeticExpression,
    RotateLeftExpression,
    RotateRightExpression,
    RangeExpression,
    QualifiedExpressionAllocation,
    SubtypeAllocation,
    IdentityExpression,
    AbsoluteExpression,
    MatchingGreaterEqualExpression,
    MatchingEqualExpression,
    MatchingUnequalExpression,
    MatchingLessThanExpression,
    MatchingLessEqualExpression,
    MatchingGreaterThanExpression,
)
from pyGHDL.dom.Concurrent import (
    ConcurrentBlockStatement,
    EntityInstantiation,
    ConfigurationInstantiation,
    ComponentInstantiation,
    ProcessStatement,
    IfGenerateStatement,
    ForGenerateStatement,
    CaseGenerateStatement,
    ConcurrentSimpleSignalAssignment,
    ConcurrentProcedureCall,
    GenericAssociationItem,
    PortAssociationItem,
    ParameterAssociationItem,
    ConcurrentAssertStatement,
)
from pyGHDL.dom.Subprogram import Function, Procedure
from pyGHDL.dom.Misc import Alias
from pyGHDL.dom.PSL import DefaultClock


@export
def GetName(node: Iir) -> Name:
    """
    Translates a *name* IIR node to a :class:`~pyVHDLModel.Name.Name`.

    A name can be a *simple name* or complex name consisting of *selected*, *parenthesis*, *attribute*, ... *names*.
    Complex names are composed as a chain of name objects, which can be traversed using the
    :attr:`~pyVHDLModel.Name.Name.Prefix` property.

    :param node: The name IIR node to translate.
    :return:     A name object.
    """
    kind = GetIirKindOfNode(node)
    if kind == nodes.Iir_Kind.Simple_Name:
        name = GetNameOfNode(node)
        return SimpleName(node, name)
    elif kind == nodes.Iir_Kind.Selected_Name:
        name = GetNameOfNode(node)
        prefixName = GetName(nodes.Get_Prefix(node))
        return SelectedName(node, name, prefixName)
    elif kind == nodes.Iir_Kind.Parenthesis_Name:
        prefixName = GetName(nodes.Get_Prefix(node))
        associations = GetAssociations(node)
        return ParenthesisName(node, prefixName, associations)
    elif kind == nodes.Iir_Kind.Attribute_Name:
        name = GetNameOfNode(node)
        prefixName = GetName(nodes.Get_Prefix(node))
        return AttributeName(node, name, prefixName)
    elif kind == nodes.Iir_Kind.Selected_By_All_Name:
        prefixName = GetName(nodes.Get_Prefix(node))
        return AllName(node, prefixName)
    else:
        raise DOMException(f"Unknown name kind '{kind.name}'")


def GetAssociations(node: Iir) -> List:
    associations = []
    for item in utils.chain_iter(nodes.Get_Association_Chain(node)):
        kind = GetIirKindOfNode(item)

        if kind in (nodes.Iir_Kind.Association_Element_By_Expression, nodes.Iir_Kind.Association_Element_By_Name):
            actual = nodes.Get_Actual(item)
            expr = GetExpressionFromNode(actual)

            associations.append(expr)
        else:
            raise DOMException(
                f"Unknown association kind '{kind.name}' in array index/slice or function call '{node}'."
            )

    return associations


@export
def GetArrayConstraintsFromSubtypeIndication(
    subtypeIndication: Iir,
) -> List:
    constraints = []
    for constraint in utils.flist_iter(nodes.Get_Index_Constraint_List(subtypeIndication)):
        constraintKind = GetIirKindOfNode(constraint)
        if constraintKind == nodes.Iir_Kind.Range_Expression:
            constraints.append(RangeExpression.parse(constraint))
        elif constraintKind in (
            nodes.Iir_Kind.Simple_Name,
            nodes.Iir_Kind.Parenthesis_Name,
            nodes.Iir_Kind.Selected_Name,
            nodes.Iir_Kind.Attribute_Name,
        ):
            constraints.append(GetName(constraint))
        else:
            position = Position.parse(constraint)
            raise DOMException(
                f"Unknown constraint kind '{constraintKind.name}' for constraint '{constraint}' in subtype indication '{subtypeIndication}' at {position}."
            )

    return constraints


@export
def GetRecordConstraintsFromSubtypeIndication(
    subtypeIndication: Iir,
) -> Dict:
    constraints = {}
    recordElementConstraint = nodes.Get_Owned_Elements_Chain(subtypeIndication)
    while recordElementConstraint != nodes.Null_Iir:
        recordElementKind = GetIirKindOfNode(recordElementConstraint)
        if recordElementKind == nodes.Iir_Kind.Record_Element_Constraint:
            recordElementName = GetNameOfNode(recordElementConstraint)
            sym = RecordElementSymbol(recordElementConstraint, SimpleName(recordElementConstraint, recordElementName))

            constraints[sym] = None
        else:
            position = Position.parse(recordElementConstraint)
            raise DOMException(
                f"Unknown constraint kind '{recordElementKind.name}' for constraint '{recordElementConstraint}' in subtype indication '{subtypeIndication}' at {position}."
            )

        recordElementConstraint = nodes.Get_Chain(recordElementConstraint)

    return constraints


@export
def GetTypeFromNode(node: Iir) -> BaseType:
    typeName = GetNameOfNode(node)
    typeDefinition = nodes.Get_Type_Definition(node)
    if typeDefinition is nodes.Null_Iir:
        return IncompleteType(node, typeName)

    kind = GetIirKindOfNode(typeDefinition)
    if kind == nodes.Iir_Kind.Enumeration_Type_Definition:
        return EnumeratedType.parse(typeName, typeDefinition)
    elif kind == nodes.Iir_Kind.Array_Type_Definition:
        return ArrayType.parse(typeName, typeDefinition)
    elif kind == nodes.Iir_Kind.Record_Type_Definition:
        return RecordType.parse(typeName, typeDefinition)
    elif kind == nodes.Iir_Kind.Access_Type_Definition:
        return AccessType.parse(typeName, typeDefinition)
    elif kind == nodes.Iir_Kind.File_Type_Definition:
        return FileType.parse(typeName, typeDefinition)
    elif kind == nodes.Iir_Kind.Protected_Type_Declaration:
        return ProtectedType.parse(typeName, typeDefinition)
    else:
        position = Position.parse(typeDefinition)
        raise DOMException(
            f"GetTypeFromNode: Unknown type definition kind '{kind.name}' for type '{typeName}' at {position}."
        )


@export
def GetAnonymousTypeFromNode(node: Iir) -> BaseType:
    typeName = GetNameOfNode(node)
    typeDefinition = nodes.Get_Type_Definition(node)
    if typeDefinition is nodes.Null_Iir:
        return IncompleteType(node, typeName)

    kind = GetIirKindOfNode(typeDefinition)
    if kind == nodes.Iir_Kind.Range_Expression:
        r = GetRangeFromNode(typeDefinition)
        return IntegerType(node, typeName, r)

    elif kind in (nodes.Iir_Kind.Attribute_Name, nodes.Iir_Kind.Parenthesis_Name):
        n = GetName(typeDefinition)

        return IntegerType(node, typeName, n)
    elif kind == nodes.Iir_Kind.Physical_Type_Definition:
        return PhysicalType.parse(typeName, typeDefinition)

    elif kind == nodes.Iir_Kind.Array_Subtype_Definition:
        WarningCollector.Raise("[NOT IMPLEMENTED] Array_Subtype_Definition")

        return ArrayType(typeDefinition, "????", [], None)
    else:
        position = Position.parse(typeDefinition)
        raise DOMException(
            f"GetAnonymousTypeFromNode: Unknown type definition kind '{kind.name}' for type '{typeName}' at {position}."
        )


@export
def GetSubtypeIndicationFromNode(node: Iir, entity: str, name: str) -> Symbol:
    """
    Extracts the subtype indication from an IIR node and translates it to a :class:`~pyVHDLModel.Symbol.Symbol`.

    .. note::

       This function calls :func:`GetSubtypeIndicationFromIndicationNode`.

       This function should be used directly if :pycode:`nodes.Get_Subtype_Indication(...)` was already called or a
       subtype indication is return by another getter.

    :param node:   The IIR node to get the subtype indication from.
    :param entity: The entity kind the subtype indication is extracted from (e.g. ``constant``). Used in exception messages.
    :param name:   The entity's name the subtype indication is extracted from (e.g. ``BITS``). Used in exception messages.
    :returns:      A symbol representing a reference to a type/subtype and possible array or record constraints.
    """
    subtypeIndicationNode = nodes.Get_Subtype_Indication(node)
    return GetSubtypeIndicationFromIndicationNode(subtypeIndicationNode, entity, name)


@export
def GetSubtypeIndicationFromIndicationNode(subtypeIndicationNode: Iir, entity: str, name: str) -> Symbol:
    """
    Translate a subtype indication to a :class:`~pyVHDLModel.Symbol.Symbol`.

    :param subtypeIndicationNode: The subtype indication IIR node.
    :param entity: The entity kind the subtype indication is extracted from (e.g. ``constant``). Used in exception messages.
    :param name:   The entity's name the subtype indication is extracted from (e.g. ``BITS``). Used in exception messages.
    :returns:      A symbol representing a reference to a type/subtype and possible array or record constraints.
    """
    if subtypeIndicationNode is nodes.Null_Iir:
        raise ValueError("Parameter 'subtypeIndicationNode' is 'Null_Iir'.")

    kind = GetIirKindOfNode(subtypeIndicationNode)
    if kind in (
        nodes.Iir_Kind.Simple_Name,
        nodes.Iir_Kind.Selected_Name,
        nodes.Iir_Kind.Attribute_Name,
    ):
        return GetSimpleTypeFromNode(subtypeIndicationNode)
    elif kind == nodes.Iir_Kind.Subtype_Definition:
        return GetScalarConstrainedSubtypeFromNode(subtypeIndicationNode)
    elif kind == nodes.Iir_Kind.Array_Subtype_Definition:
        return GetArrayConstrainedSubtypeFromNode(subtypeIndicationNode)
    elif kind == nodes.Iir_Kind.Record_Subtype_Definition:
        return GetRecordConstrainedSubtypeFromNode(subtypeIndicationNode)
    else:
        raise DOMException(f"Unknown kind '{kind.name}' for an subtype indication in a {entity} of `{name}`.")


@export
def GetSimpleTypeFromNode(subtypeIndicationNode: Iir) -> SimpleSubtypeSymbol:
    subtypeName = GetName(subtypeIndicationNode)
    return SimpleSubtypeSymbol(subtypeIndicationNode, subtypeName)


@export
def GetScalarConstrainedSubtypeFromNode(
    subtypeIndicationNode: Iir,
) -> ConstrainedScalarSubtypeSymbol:
    typeMark = nodes.Get_Subtype_Type_Mark(subtypeIndicationNode)
    typeMarkName = GetNameOfNode(typeMark)
    simpleTypeMark = SimpleName(typeMark, typeMarkName)
    rangeConstraint = nodes.Get_Range_Constraint(subtypeIndicationNode)

    r = None
    # Check if RangeExpression. Might also be an AttributeName (see §3.1)
    if rangeConstraint != nodes.Null_Iir:
        if GetIirKindOfNode(rangeConstraint) == nodes.Iir_Kind.Range_Expression:
            r = GetRangeFromNode(rangeConstraint)
    # TODO: Get actual range from AttributeName node?

    return ConstrainedScalarSubtypeSymbol(subtypeIndicationNode, simpleTypeMark, r)


@export
def GetArrayConstrainedSubtypeFromNode(
    subtypeIndicationNode: Iir,
) -> ConstrainedArraySubtypeSymbol:
    typeMark = nodes.Get_Subtype_Type_Mark(subtypeIndicationNode)
    typeMarkName = GetNameOfNode(typeMark)
    simpleTypeMark = SimpleName(typeMark, typeMarkName)

    constraints = GetArrayConstraintsFromSubtypeIndication(subtypeIndicationNode)
    return ConstrainedArraySubtypeSymbol(subtypeIndicationNode, simpleTypeMark, constraints)


@export
def GetRecordConstrainedSubtypeFromNode(
    subtypeIndicationNode: Iir,
) -> ConstrainedRecordSubtypeSymbol:
    typeMark = nodes.Get_Subtype_Type_Mark(subtypeIndicationNode)
    typeMarkName = GetNameOfNode(typeMark)
    simpleTypeMark = SimpleName(typeMark, typeMarkName)

    constraints = GetRecordConstraintsFromSubtypeIndication(subtypeIndicationNode)
    return ConstrainedRecordSubtypeSymbol(subtypeIndicationNode, simpleTypeMark, constraints)


@export
def GetSubtypeFromNode(subtypeNode: Iir) -> Symbol:
    subtypeName = GetNameOfNode(subtypeNode)
    symbol = GetSubtypeIndicationFromNode(subtypeNode, "subtype", subtypeName)

    return Subtype(subtypeNode, subtypeName, symbol)


@export
def GetRangeFromNode(node: Iir) -> Range:
    """
    Translate a range IIR node to a :class:`~pyVHDLModel.Base.Range`.

    :param node: The IIR node representing a range.
    :return:     The translated range object.
    """
    direction = nodes.Get_Direction(node)
    leftBound = nodes.Get_Left_Limit_Expr(node)
    rightBound = nodes.Get_Right_Limit_Expr(node)

    return Range(
        GetExpressionFromNode(leftBound),
        GetExpressionFromNode(rightBound),
        Direction.DownTo if direction else Direction.To,
    )


__EXPRESSION_TRANSLATION = {
    nodes.Iir_Kind.Simple_Name: SimpleObjectOrFunctionCallSymbol,
    nodes.Iir_Kind.Selected_Name: IndexedObjectOrFunctionCallSymbol,
    nodes.Iir_Kind.Attribute_Name: IndexedObjectOrFunctionCallSymbol,
    nodes.Iir_Kind.Parenthesis_Name: IndexedObjectOrFunctionCallSymbol,
    nodes.Iir_Kind.Null_Literal: NullLiteral,
    nodes.Iir_Kind.Integer_Literal: IntegerLiteral,
    nodes.Iir_Kind.Floating_Point_Literal: FloatingPointLiteral,
    nodes.Iir_Kind.Physical_Int_Literal: PhysicalIntegerLiteral,
    nodes.Iir_Kind.Physical_Fp_Literal: PhysicalFloatingLiteral,
    nodes.Iir_Kind.Character_Literal: CharacterLiteral,
    nodes.Iir_Kind.String_Literal8: StringLiteral,
    nodes.Iir_Kind.Identity_Operator: IdentityExpression,
    nodes.Iir_Kind.Negation_Operator: NegationExpression,
    nodes.Iir_Kind.Absolute_Operator: AbsoluteExpression,
    nodes.Iir_Kind.Range_Expression: RangeExpression,
    nodes.Iir_Kind.Addition_Operator: AdditionExpression,
    nodes.Iir_Kind.Concatenation_Operator: ConcatenationExpression,
    nodes.Iir_Kind.Not_Operator: InverseExpression,
    nodes.Iir_Kind.Parenthesis_Expression: ParenthesisExpression,
    nodes.Iir_Kind.Substraction_Operator: SubtractionExpression,
    nodes.Iir_Kind.Multiplication_Operator: MultiplyExpression,
    nodes.Iir_Kind.Division_Operator: DivisionExpression,
    nodes.Iir_Kind.Modulus_Operator: ModuloExpression,
    nodes.Iir_Kind.Remainder_Operator: RemainderExpression,
    nodes.Iir_Kind.Exponentiation_Operator: ExponentiationExpression,
    nodes.Iir_Kind.And_Operator: AndExpression,
    nodes.Iir_Kind.Nand_Operator: NandExpression,
    nodes.Iir_Kind.Or_Operator: OrExpression,
    nodes.Iir_Kind.Nor_Operator: NorExpression,
    nodes.Iir_Kind.Xor_Operator: XorExpression,
    nodes.Iir_Kind.Xnor_Operator: XnorExpression,
    nodes.Iir_Kind.Equality_Operator: EqualExpression,
    nodes.Iir_Kind.Inequality_Operator: UnequalExpression,
    nodes.Iir_Kind.Less_Than_Operator: LessThanExpression,
    nodes.Iir_Kind.Less_Than_Or_Equal_Operator: LessEqualExpression,
    nodes.Iir_Kind.Greater_Than_Operator: GreaterThanExpression,
    nodes.Iir_Kind.Greater_Than_Or_Equal_Operator: GreaterEqualExpression,
    nodes.Iir_Kind.Match_Equality_Operator: MatchingEqualExpression,
    nodes.Iir_Kind.Match_Inequality_Operator: MatchingUnequalExpression,
    nodes.Iir_Kind.Match_Less_Than_Operator: MatchingLessThanExpression,
    nodes.Iir_Kind.Match_Less_Than_Or_Equal_Operator: MatchingLessEqualExpression,
    nodes.Iir_Kind.Match_Greater_Than_Operator: MatchingGreaterThanExpression,
    nodes.Iir_Kind.Match_Greater_Than_Or_Equal_Operator: MatchingGreaterEqualExpression,
    nodes.Iir_Kind.Sll_Operator: ShiftLeftLogicExpression,
    nodes.Iir_Kind.Srl_Operator: ShiftRightLogicExpression,
    nodes.Iir_Kind.Sla_Operator: ShiftLeftArithmeticExpression,
    nodes.Iir_Kind.Sra_Operator: ShiftRightArithmeticExpression,
    nodes.Iir_Kind.Rol_Operator: RotateLeftExpression,
    nodes.Iir_Kind.Ror_Operator: RotateRightExpression,
    nodes.Iir_Kind.Qualified_Expression: QualifiedExpression,
    nodes.Iir_Kind.Aggregate: Aggregate,
    nodes.Iir_Kind.Allocator_By_Subtype: SubtypeAllocation,
    nodes.Iir_Kind.Allocator_By_Expression: QualifiedExpressionAllocation,
}


@export
def GetExpressionFromNode(node: Iir) -> ExpressionUnion:
    """
    Translates an expression IIR node to an :class:`~pyVHDLModel.Expression.BaseExpression`.

    :param node: The IIR node representing an expression.
    :return:     The translated expression.
    """
    kind = GetIirKindOfNode(node)

    try:
        cls = __EXPRESSION_TRANSLATION[kind]
    except KeyError:
        position = Position.parse(node)
        raise DOMException(f"Unknown expression kind '{kind.name}' in expression '{node}' at {position}.")

    return cls.parse(node)


@export
def GetGenericsFromChainedNodes(nodeChain: Iir) -> Generator[GenericInterfaceItemMixin, None, None]:
    """
    Translates a chain of generics (IIR nodes) to a sequence of :class:`pyVHDLModel.Interface.GenericInterfaceItem`.

    :param nodeChain: The IIR node representing the first generic in the chain.
    :return:          A generator returning generic interface items.
    """
    from pyGHDL.dom.InterfaceItem import (
        GenericTypeInterfaceItem,
        GenericPackageInterfaceItem,
        GenericProcedureInterfaceItem,
        GenericFunctionInterfaceItem,
    )

    furtherIdentifiers = []
    generic = nodeChain
    while generic != nodes.Null_Iir:
        kind = GetIirKindOfNode(generic)
        if kind == nodes.Iir_Kind.Interface_Constant_Declaration:
            from pyGHDL.dom.InterfaceItem import GenericConstantInterfaceItem

            parseNode = generic

            # Lookahead for generics with multiple identifiers at once
            if nodes.Get_Has_Identifier_List(generic):
                nextNode = nodes.Get_Chain(generic)
                for nextGeneric in utils.chain_iter(nextNode):
                    # Consecutive identifiers are found, if the subtype indication is Null
                    if nodes.Get_Subtype_Indication(nextGeneric) == nodes.Null_Iir:
                        furtherIdentifiers.append(GetNameOfNode(nextGeneric))
                    else:
                        generic = nextGeneric
                        break

                    # The last consecutive identifiers has no Identifier_List flag
                    if not nodes.Get_Has_Identifier_List(nextGeneric):
                        generic = nodes.Get_Chain(nextGeneric)
                        break
                else:
                    generic = nodes.Null_Iir
            else:
                generic = nodes.Get_Chain(generic)

            yield GenericConstantInterfaceItem.parse(parseNode, furtherIdentifiers)
            furtherIdentifiers.clear()
            continue
        else:
            if kind == nodes.Iir_Kind.Interface_Type_Declaration:
                yield GenericTypeInterfaceItem.parse(generic)
            elif kind == nodes.Iir_Kind.Interface_Package_Declaration:
                yield GenericPackageInterfaceItem.parse(generic)
            elif kind == nodes.Iir_Kind.Interface_Procedure_Declaration:
                yield GenericProcedureInterfaceItem.parse(generic)
            elif kind == nodes.Iir_Kind.Interface_Function_Declaration:
                yield GenericFunctionInterfaceItem.parse(generic)
            else:
                position = Position.parse(generic)
                raise DOMException(f"Unknown generic kind '{kind.name}' in generic '{generic}' at {position}.")

        generic = nodes.Get_Chain(generic)


@export
def GetPortsFromChainedNodes(nodeChain: Iir) -> Generator[PortInterfaceItemMixin, None, None]:
    """
    Translates a chain of ports (IIR nodes) to a sequence of :class:`pyVHDLModel.Interface.PortInterfaceItem`.

    :param nodeChain: The IIR node representing the first port in the chain.
    :return:          A generator returning port interface items.
    """
    furtherIdentifiers = []
    port = nodeChain
    while port != nodes.Null_Iir:
        kind = GetIirKindOfNode(port)
        if kind == nodes.Iir_Kind.Interface_Signal_Declaration:
            from pyGHDL.dom.InterfaceItem import PortSignalInterfaceItem

            portToParse = port

            # Lookahead for ports with multiple identifiers at once
            if nodes.Get_Has_Identifier_List(port):
                nextNode = nodes.Get_Chain(port)
                for nextPort in utils.chain_iter(nextNode):
                    # Consecutive identifiers are found, if the subtype indication is Null
                    if nodes.Get_Subtype_Indication(nextPort) == nodes.Null_Iir:
                        furtherIdentifiers.append(GetNameOfNode(nextPort))
                    else:
                        port = nextPort
                        break

                    # The last consecutive identifiers has no Identifier_List flag
                    if not nodes.Get_Has_Identifier_List(nextPort):
                        port = nodes.Get_Chain(nextPort)
                        break
                else:
                    port = nodes.Null_Iir
            else:
                port = nodes.Get_Chain(port)

            yield PortSignalInterfaceItem.parse(portToParse, furtherIdentifiers)
            furtherIdentifiers.clear()
            continue
        else:
            position = Position.parse(port)
            raise DOMException(f"Unknown port kind '{kind.name}' in port '{port}' at {position}.")


@export
def GetParameterFromChainedNodes(nodeChain: Iir) -> Generator[ParameterInterfaceItemMixin, None, None]:
    """
    Translates a chain of parameters (IIR nodes) to a sequence of :class:`pyVHDLModel.Interface.ParameterInterfaceItem`.

    :param nodeChain: The IIR node representing the first parameter in the chain.
    :return:          A generator returning parameter interface items.
    """
    identifiers = []
    parameter = nodeChain
    while parameter != nodes.Null_Iir:
        kind = GetIirKindOfNode(parameter)
        if kind == nodes.Iir_Kind.Interface_Constant_Declaration:
            from pyGHDL.dom.InterfaceItem import ParameterConstantInterfaceItem

            parseMethod = ParameterConstantInterfaceItem.parse
            parseNode = parameter
        elif kind == nodes.Iir_Kind.Interface_Variable_Declaration:
            from pyGHDL.dom.InterfaceItem import ParameterVariableInterfaceItem

            parseMethod = ParameterVariableInterfaceItem.parse
            parseNode = parameter
        elif kind == nodes.Iir_Kind.Interface_Signal_Declaration:
            from pyGHDL.dom.InterfaceItem import ParameterSignalInterfaceItem

            parseMethod = ParameterSignalInterfaceItem.parse
            parseNode = parameter
        elif kind == nodes.Iir_Kind.Interface_File_Declaration:
            from pyGHDL.dom.InterfaceItem import ParameterFileInterfaceItem

            parseMethod = ParameterFileInterfaceItem.parse
            parseNode = parameter
        else:
            position = Position.parse(parameter)
            raise DOMException(f"Unknown parameter kind '{kind.name}' in parameter '{parameter}' at {position}.")

        # Lookahead for parameters with multiple identifiers at once
        if nodes.Get_Has_Identifier_List(parameter):
            nextNode = nodes.Get_Chain(parameter)
            for nextParameter in utils.chain_iter(nextNode):
                # Consecutive identifiers are found, if the subtype indication is Null
                if nodes.Get_Subtype_Indication(nextParameter) == nodes.Null_Iir:
                    identifiers.append(GetNameOfNode(nextParameter))
                else:
                    parameter = nextParameter
                    break

                # The last consecutive identifiers has no Identifier_List flag
                if not nodes.Get_Has_Identifier_List(nextParameter):
                    parameter = nodes.Get_Chain(nextParameter)
                    break
            else:
                parameter = nodes.Null_Iir
        else:
            parameter = nodes.Get_Chain(parameter)

        yield parseMethod(parseNode, identifiers)


def GetMapAspect(mapAspect: Iir, cls: Type, entity: str) -> Generator[AssociationItem, None, None]:
    for item in utils.chain_iter(mapAspect):
        kind = GetIirKindOfNode(item)
        if kind is nodes.Iir_Kind.Association_Element_By_Expression:
            formalNode = nodes.Get_Formal(item)
            if formalNode is nodes.Null_Iir:
                formal = None
            else:
                formal = GetName(formalNode)

            actual = GetExpressionFromNode(nodes.Get_Actual(item))

            yield cls(item, actual, formal)
        elif kind is nodes.Iir_Kind.Association_Element_Open:
            formalNode = nodes.Get_Formal(item)
            if formalNode is nodes.Null_Iir:
                formal = None
            else:
                formal = GetName(formalNode)

            yield cls(item, OpenName(item), formal)
        else:
            pos = Position.parse(item)
            raise DOMException(f"Unknown association kind '{kind.name}' in {entity} map at line {pos.Line}.")


def GetGenericMapAspect(genericMapAspect: Iir) -> Generator[GenericAssociationItem, None, None]:
    return GetMapAspect(genericMapAspect, GenericAssociationItem, "generic")


def GetPortMapAspect(portMapAspect: Iir) -> Generator[PortAssociationItem, None, None]:
    return GetMapAspect(portMapAspect, PortAssociationItem, "port")


def GetParameterMapAspect(parameterMapAspect: Iir) -> Generator[ParameterAssociationItem, None, None]:
    return GetMapAspect(parameterMapAspect, ParameterAssociationItem, "parameter")


def GetDeclaredItemsFromChainedNodes(nodeChain: Iir, entity: str, name: str) -> Generator[ModelEntity, None, None]:
    furtherIdentifiers = []
    item = nodeChain
    lastKind = None
    while item != nodes.Null_Iir:
        kind = GetIirKindOfNode(item)
        if kind == nodes.Iir_Kind.Constant_Declaration:
            from pyGHDL.dom.Object import Constant

            objectParseMethod = Constant.parse
            objectItem = item
        elif kind == nodes.Iir_Kind.Variable_Declaration:
            from pyGHDL.dom.Object import SharedVariable

            if nodes.Get_Shared_Flag(item):
                objectParseMethod = SharedVariable.parse
                objectItem = item
            else:
                objectParseMethod = Variable.parse
                objectItem = item
        elif kind == nodes.Iir_Kind.Signal_Declaration:
            from pyGHDL.dom.Object import Signal

            objectParseMethod = Signal.parse
            objectItem = item
        elif kind == nodes.Iir_Kind.File_Declaration:
            from pyGHDL.dom.Object import File

            objectParseMethod = File.parse
            objectItem = item
        else:
            if kind == nodes.Iir_Kind.Type_Declaration:
                yield GetTypeFromNode(item)

            elif kind == nodes.Iir_Kind.Anonymous_Type_Declaration:
                yield GetAnonymousTypeFromNode(item)

            elif kind == nodes.Iir_Kind.Subtype_Declaration:
                yield GetSubtypeFromNode(item)

            elif kind == nodes.Iir_Kind.Function_Declaration:
                if nodes.Get_Has_Body(item):
                    yield Function.parse(item)
                else:
                    WarningCollector.Raise("[NOT IMPLEMENTED] function declaration without body")

                lastKind = kind
                item = nodes.Get_Chain(item)
                continue
            elif kind == nodes.Iir_Kind.Function_Body:
                if lastKind is nodes.Iir_Kind.Function_Declaration:
                    pass
                else:
                    position = Position.parse(item)
                    raise DOMException(
                        f"Found unexpected function body '{GetNameOfNode(item)}' in {entity} '{name}' at {position}."
                    )
            elif kind == nodes.Iir_Kind.Procedure_Declaration:
                if nodes.Get_Has_Body(item):
                    yield Procedure.parse(item)
                else:
                    WarningCollector.Raise("[NOT IMPLEMENTED] procedure declaration without body")

                lastKind = kind
                item = nodes.Get_Chain(item)
                continue
            elif kind == nodes.Iir_Kind.Procedure_Body:
                if lastKind is nodes.Iir_Kind.Procedure_Declaration:
                    pass
                else:
                    position = Position.parse(item)
                    raise DOMException(
                        f"Found unexpected procedure body '{GetNameOfNode(item)}' in {entity} '{name}' at {position}."
                    )
            elif kind == nodes.Iir_Kind.Protected_Type_Body:
                yield ProtectedTypeBody.parse(item)
            elif kind == nodes.Iir_Kind.Object_Alias_Declaration:
                yield GetAliasFromNode(item)
            elif kind == nodes.Iir_Kind.Component_Declaration:
                from pyGHDL.dom.DesignUnit import Component

                yield Component.parse(item)
            elif kind == nodes.Iir_Kind.Attribute_Declaration:
                from pyGHDL.dom.Attribute import Attribute

                yield Attribute.parse(item)
            elif kind == nodes.Iir_Kind.Attribute_Specification:
                from pyGHDL.dom.Attribute import AttributeSpecification

                yield AttributeSpecification.parse(item)
            elif kind == nodes.Iir_Kind.Use_Clause:
                from pyGHDL.dom.DesignUnit import UseClause

                yield UseClause.parse(item)
            elif kind == nodes.Iir_Kind.Package_Declaration:
                from pyGHDL.dom.DesignUnit import Package

                yield Package.parse(item, None)  # TODO: Can it have a context?
            elif kind == nodes.Iir_Kind.Package_Instantiation_Declaration:
                from pyGHDL.dom.DesignUnit import PackageInstantiation

                yield PackageInstantiation.parse(item)
            elif kind == nodes.Iir_Kind.Configuration_Specification:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Configuration specification in {name}")
            elif kind == nodes.Iir_Kind.Psl_Default_Clock:
                yield DefaultClock.parse(item)
            elif kind == nodes.Iir_Kind.Group_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Group declaration in {name}")
            elif kind == nodes.Iir_Kind.Group_Template_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Group template declaration in {name}")
            elif kind == nodes.Iir_Kind.Disconnection_Specification:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Disconnect specification in {name}")
            elif kind == nodes.Iir_Kind.Nature_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Nature declaration in {name}")
            elif kind == nodes.Iir_Kind.Free_Quantity_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Free quantity declaration in {name}")
            elif kind == nodes.Iir_Kind.Across_Quantity_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Across quantity declaration in {name}")
            elif kind == nodes.Iir_Kind.Through_Quantity_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Through quantity declaration in {name}")
            elif kind == nodes.Iir_Kind.Terminal_Declaration:
                WarningCollector.Raise(f"[NOT IMPLEMENTED] Terminal declaration in {name}")
            else:
                position = Position.parse(item)
                raise DOMException(f"Unknown declared item kind '{kind.name}' in {entity} '{name}' at {position}.")

            lastKind = None
            item = nodes.Get_Chain(item)
            continue

        # Lookahead for objects with multiple identifiers at once
        if nodes.Get_Has_Identifier_List(item):
            nextNode = nodes.Get_Chain(item)
            for nextItem in utils.chain_iter(nextNode):
                # Consecutive identifiers are found, if the subtype indication is Null
                if nodes.Get_Subtype_Indication(nextItem) == nodes.Null_Iir:
                    furtherIdentifiers.append(GetNameOfNode(nextItem))
                else:
                    item = nextItem
                    break

                # The last consecutive identifiers has no Identifier_List flag
                if not nodes.Get_Has_Identifier_List(nextItem):
                    item = nodes.Get_Chain(nextItem)
                    break
            else:
                item = nodes.Null_Iir
        else:
            item = nodes.Get_Chain(item)

        yield objectParseMethod(objectItem, furtherIdentifiers)
        furtherIdentifiers.clear()


def GetConcurrentStatementsFromChainedNodes(
    nodeChain: Iir, entity: str, name: str
) -> Generator[ConcurrentStatement, None, None]:
    for statement in utils.chain_iter(nodeChain):
        label = nodes.Get_Label(statement)
        label = name_table.Get_Name_Ptr(label) if label != nodes.Null_Iir else None

        position = Position.parse(statement)

        kind = GetIirKindOfNode(statement)
        if kind == nodes.Iir_Kind.Sensitized_Process_Statement:
            yield ProcessStatement.parse(statement, label, True)

        elif kind == nodes.Iir_Kind.Process_Statement:
            yield ProcessStatement.parse(statement, label, False)

        elif kind == nodes.Iir_Kind.Concurrent_Simple_Signal_Assignment:
            yield ConcurrentSimpleSignalAssignment.parse(statement, label)
        elif kind == nodes.Iir_Kind.Concurrent_Conditional_Signal_Assignment:
            WarningCollector.Raise(
                f"[NOT IMPLEMENTED] Concurrent (conditional) signal assignment (label: '{label}') at line {position.Line}"
            )
        elif kind == nodes.Iir_Kind.Concurrent_Selected_Signal_Assignment:
            WarningCollector.Raise(
                f"[NOT IMPLEMENTED] Concurrent (selected) signal assignment (label: '{label}') at line {position.Line}"
            )
        elif kind == nodes.Iir_Kind.Concurrent_Procedure_Call_Statement:
            yield ConcurrentProcedureCall.parse(statement, label)
        elif kind == nodes.Iir_Kind.Component_Instantiation_Statement:
            instantiatedUnit = nodes.Get_Instantiated_Unit(statement)
            instantiatedUnitKind = GetIirKindOfNode(instantiatedUnit)

            if instantiatedUnitKind == nodes.Iir_Kind.Entity_Aspect_Entity:
                yield EntityInstantiation.parse(statement, instantiatedUnit, label)
            elif instantiatedUnitKind == nodes.Iir_Kind.Entity_Aspect_Configuration:
                yield ConfigurationInstantiation.parse(statement, instantiatedUnit, label)
            elif (
                instantiatedUnitKind == nodes.Iir_Kind.Simple_Name
                or instantiatedUnitKind == nodes.Iir_Kind.Selected_Name
            ):
                yield ComponentInstantiation.parse(statement, instantiatedUnit, label)
            else:
                raise DOMException(
                    f"Unknown instantiation kind '{instantiatedUnitKind.name}' in instantiation of label {label} at {position}."
                )
        elif kind == nodes.Iir_Kind.Block_Statement:
            yield ConcurrentBlockStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.If_Generate_Statement:
            yield IfGenerateStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Case_Generate_Statement:
            yield CaseGenerateStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.For_Generate_Statement:
            yield ForGenerateStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Psl_Assert_Directive:
            yield ConcurrentAssertStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Simple_Simultaneous_Statement:
            WarningCollector.Raise(
                f"[NOT IMPLEMENTED] Simple simultaneous statement (label: '{label}') at line {position.Line}"
            )
        else:
            raise DOMException(f"Unknown statement of kind '{kind.name}' in {entity} '{name}' at {position}.")


def GetSequentialStatementsFromChainedNodes(
    nodeChain: Iir, entity: str, name: str
) -> Generator[SequentialStatement, None, None]:
    for statement in utils.chain_iter(nodeChain):
        label = nodes.Get_Label(statement)
        label = name_table.Get_Name_Ptr(label) if label != nodes.Null_Iir else None

        position = Position.parse(statement)
        kind = GetIirKindOfNode(statement)
        if kind == nodes.Iir_Kind.If_Statement:
            yield IfStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.For_Loop_Statement:
            yield ForLoopStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Case_Statement:
            yield CaseStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Simple_Signal_Assignment_Statement:
            yield SequentialSimpleSignalAssignment.parse(statement, label)
        elif kind in (
            nodes.Iir_Kind.Variable_Assignment_Statement,
            nodes.Iir_Kind.Conditional_Variable_Assignment_Statement,
        ):
            WarningCollector.Raise(f"[NOT IMPLEMENTED] Variable assignment (label: '{label}') at line {position.Line}")
        elif kind == nodes.Iir_Kind.Wait_Statement:
            yield WaitStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Procedure_Call_Statement:
            yield SequentialProcedureCall.parse(statement, label)
        elif kind == nodes.Iir_Kind.Report_Statement:
            yield SequentialReportStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Assertion_Statement:
            yield SequentialAssertStatement.parse(statement, label)
        elif kind == nodes.Iir_Kind.Null_Statement:
            yield NullStatement(statement, label)
        elif kind == nodes.Iir_Kind.Exit_Statement:
            yield ExitStatement(statement, label)
        else:
            raise DOMException(f"Unknown statement of kind '{kind.name}' in {entity} '{name}' at {position}.")


def GetAliasFromNode(aliasNode: Iir):
    aliasName = GetNameOfNode(aliasNode)

    return Alias(aliasNode, aliasName)



from os.path import dirname, join
import json
import re

# Try to load JSON data from a file. If not found, use the argument as a tag name and retrieve the data from GitHub.
def getJSON(tag='all'):
   f = tag
   tag = '/'+tag
   if f == 'all':
      f = 'releases'
      tag = ''

   try:
      d = json.loads(open(join(dirname(__file__), f+'.json'), 'r').read())
   except:
      from urllib.request import urlopen
      d = json.loads(urlopen('https://api.github.com/repos/ghdl/ghdl/releases'+tag).read())
      json.dump(d, open(f+'.json', 'w'), indent=4)
   return d

#
# Print two versions of each shield. One for 'html' (`image::`) and one for 'latex' (`replace::`)
#

def printShieldSrc(label, alt, img, target, latex=False):
   if latex:
      if label[-6:] == '/total':
         label = label[:-6]
      idx = re.compile(r"[\W_]+").sub('', label)
      print(f"""
.. |l{idx}| replace:: `{label}`_
.. _{label}: {target}
""")
   else:
      print(f'''
.. |{label}| image:: {img}
   :target: {target}
   :height: 22
   :alt: {alt}
   :class: shield
''')

#
# Create shields/badges from JSON file
#

def createShields(file='shields'):
   shields = getJSON(file)
   for k, v in shields.items():
      t = v['target']
      if t[0:3] == 'gh:':
        t = 'https://github.com/' + t[3:]

      printShieldSrc(
         'SHIELD:'+k,
         v['alt'],
         'https://img.shields.io/' + v['src'] + '&style=flat-square&longCache=true',
         t,
         False
      )



import cocotb
from cocotb.triggers import Timer
from cocotb.binary import BinaryValue

@cocotb.test()
async def test_deposit_bug(dut):
    dut.o.value = 0 # This is a deposit for Cocotb
    await Timer(1, 'ns')

    # From this point, the deposit is ignored and Cocotb has full control over the signal o
    dut.d.value = 1
    dut.oe.value = 1
    await Timer(1, 'ns')
    assert dut.o.value == 1



# Auto generated Python source file from Ada sources
# Call 'make' in 'src/vhdl' to regenerate:
#
from pyTooling.Decorators import export

from pyGHDL.libghdl import libghdl


@export
def Get_Start_Location(obj):
    return libghdl.vhdl__elocations__get_start_location(obj)


@export
def Set_Start_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_start_location(obj, value)


@export
def Get_Right_Paren_Location(obj):
    return libghdl.vhdl__elocations__get_right_paren_location(obj)


@export
def Set_Right_Paren_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_right_paren_location(obj, value)


@export
def Get_End_Location(obj):
    return libghdl.vhdl__elocations__get_end_location(obj)


@export
def Set_End_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_end_location(obj, value)


@export
def Get_Is_Location(obj):
    return libghdl.vhdl__elocations__get_is_location(obj)


@export
def Set_Is_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_is_location(obj, value)


@export
def Get_Begin_Location(obj):
    return libghdl.vhdl__elocations__get_begin_location(obj)


@export
def Set_Begin_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_begin_location(obj, value)


@export
def Get_Then_Location(obj):
    return libghdl.vhdl__elocations__get_then_location(obj)


@export
def Set_Then_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_then_location(obj, value)


@export
def Get_Use_Location(obj):
    return libghdl.vhdl__elocations__get_use_location(obj)


@export
def Set_Use_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_use_location(obj, value)


@export
def Get_Loop_Location(obj):
    return libghdl.vhdl__elocations__get_loop_location(obj)


@export
def Set_Loop_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_loop_location(obj, value)


@export
def Get_Generate_Location(obj):
    return libghdl.vhdl__elocations__get_generate_location(obj)


@export
def Set_Generate_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_generate_location(obj, value)


@export
def Get_Generic_Location(obj):
    return libghdl.vhdl__elocations__get_generic_location(obj)


@export
def Set_Generic_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_generic_location(obj, value)


@export
def Get_Port_Location(obj):
    return libghdl.vhdl__elocations__get_port_location(obj)


@export
def Set_Port_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_port_location(obj, value)


@export
def Get_Generic_Map_Location(obj):
    return libghdl.vhdl__elocations__get_generic_map_location(obj)


@export
def Set_Generic_Map_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_generic_map_location(obj, value)


@export
def Get_Port_Map_Location(obj):
    return libghdl.vhdl__elocations__get_port_map_location(obj)


@export
def Set_Port_Map_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_port_map_location(obj, value)


@export
def Get_Arrow_Location(obj):
    return libghdl.vhdl__elocations__get_arrow_location(obj)


@export
def Set_Arrow_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_arrow_location(obj, value)


@export
def Get_Colon_Location(obj):
    return libghdl.vhdl__elocations__get_colon_location(obj)


@export
def Set_Colon_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_colon_location(obj, value)


@export
def Get_Assign_Location(obj):
    return libghdl.vhdl__elocations__get_assign_location(obj)


@export
def Set_Assign_Location(obj, value) -> None:
    libghdl.vhdl__elocations__set_assign_location(obj, value)



# =============================================================================
#               ____ _   _ ____  _          _
#  _ __  _   _ / ___| | | |  _ \| |      __| | ___  _ __ ___
# | '_ \| | | | |  _| |_| | | | | |     / _` |/ _ \| '_ ` _ \
# | |_) | |_| | |_| |  _  | |_| | |___ | (_| | (_) | | | | | |
# | .__/ \__, |\____|_| |_|____/|_____(_)__,_|\___/|_| |_| |_|
# |_|    |___/
# =============================================================================
# Authors:
#   Patrick Lehmann
#
# License:
# ============================================================================
#  Copyright (C) 2019-2021 Tristan Gingold
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <gnu.org/licenses>.
#
# SPDX-License-Identifier: GPL-2.0-or-later
# ============================================================================
"""
Document object model (DOM) for :mod:`pyGHDL.libghdl` based on :doc:`pyVHDLModel <vhdlmodel:index>`.
"""
from pathlib import Path

from pyTooling.Decorators import export, readonly
from pyTooling.MetaClasses import ExtendedType

from pyGHDL import GHDLBaseException
from pyGHDL.libghdl import files_map, name_table
from pyGHDL.libghdl._types import Iir
from pyGHDL.libghdl.vhdl import nodes


@export
class Position(metaclass=ExtendedType):
    """Represents the source code position of a IIR node in a source file."""

    _filename: Path
    _line: int
    _column: int

    def __init__(self, filename: Path, line: int, column: int) -> None:
        self._filename = filename
        self._line = line
        self._column = column

    @classmethod
    def parse(cls, node: Iir) -> "Position":
        """Return the source code position of a IIR node."""
        if node == nodes.Null_Iir:
            raise ValueError("Position.parse(): Parameter 'node' must not be 'Null_iir'.")

        location = nodes.Get_Location(node)
        file = files_map.Location_To_File(location)
        fileNameId = files_map.Get_File_Name(file)
        fileName = name_table.Get_Name_Ptr(fileNameId)
        line = files_map.Location_File_To_Line(location, file)
        column = files_map.Location_File_Line_To_Offset(location, file, line)

        return cls(Path(fileName), line, column)

    @readonly
    def Filename(self) -> Path:
        """
        Read-only property to access the filename this source code position referres to (:attr:`_filename`).

        :returns: The source code position's filename.
        """
        return self._filename

    @readonly
    def Line(self) -> int:
        """
        Read-only property to access the line number this source code position referres to (:attr:`_line`).

        :returns: The source code position's line.
        """
        return self._line

    @readonly
    def Column(self) -> int:
        """
        Read-only property to access the column this source code position referres to (:attr:`_column`).

        :returns: The source code position's column.
        """
        return self._column

    def __str__(self):
        return f"{self._filename}:{self._line}:{self._column}"


@export
class DOMMixin(metaclass=ExtendedType, mixin=True):
    _iirNode: Iir
    _position: Position

    def __init__(self, node: Iir) -> None:
        self._iirNode = node
        self._position = None

    # TODO: add Iir and Iir kind properties

    @readonly
    def Position(self) -> Position:
        """
        Cached read-only property to access an IIR's position in source code (:attr:`_position`).

        If :attr:`_position` is None, resolve the position object from :attr:`_iirNode`

        :returns: The IIR's position in the source file.
        """
        if self._position is None:
            self._position = Position.parse(self._iirNode)

        return self._position


@export
class DOMException(GHDLBaseException):
    pass



#!/usr/bin/env python
#  Generate the body of ieee.numeric_std and numeric_bit from a template.
#  The implementation is based only on the specification and on testing (as
#  the specifications are often ambiguous).
#  The algorithms are very simple: carry ripple adder, restoring division.
#  This file is part of GHDL.
#  Both this file and the outputs of this file are copyrighted.
#  Copyright (C) 2015-2021 Tristan Gingold
#
#  GHDL is free software; you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 2, or (at your option) any later
#  version.
#
#  GHDL is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
#  for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with GHDL; see the file COPYING.md.  If not see
#  <http://www.gnu.org/licenses/>.

import sys

# My python 'style' and knowledge is basic...  Do not hesitate to comment.

binary_funcs = ["and", "nand", "or", "nor", "xor"]
compare_funcs = ["=", "/=", ">", ">=", "<", "<="]

vec_types = ["UNSIGNED", "SIGNED"]

logics = ["bit", "std"]
logic_types = {"bit": "bit", "std": "sl_x01"}
logic_undefs = {"bit": "'0'", "std": "'X'"}

logic = "xx"  # Current logic, either bit or std

v93 = False

# Stream to write.
out = sys.stdout


def w(s):
    """Write S to the output."""
    out.write(s)


def logic_type():
    return logic_types[logic]


def logic_undef():
    return logic_undefs[logic]


def disp_vec_binary(func, typ):
    """Generate the body of a vector binary logic function."""
    res = (
        """
  function "{0}" (l, r : {1}) return {1}
  is
    subtype res_type is {1} (l'length - 1 downto 0);
    alias la : res_type is l;
    alias ra : {1} (r'length - 1 downto 0) is r;
    variable res : res_type;
  begin
    if la'left /= ra'left then
      assert false
        report "NUMERIC_STD.""{0}"": arguments are not of the same length"
        severity failure;
      res := (others => """
        + logic_undef()
        + """);
    else
      for I in res_type'range loop
        res (I) := la (I) {0} ra (I);
      end loop;
    end if;
    return res;
  end "{0}";\n"""
    )
    w(res.format(func, typ))


def disp_non_logical_warning(func):
    return """
        assert NO_WARNING
          report "NUMERIC_STD.""{0}"": non logical value detected"
          severity warning;""".format(
        func
    )


def conv_bit(expr):
    if logic == "std":
        return "sl_to_x01 (" + expr + ")"
    else:
        return expr


def extract_bit(name):
    res = "{0}b := " + conv_bit("{0}a (i)") + ";"
    return res.format(name)


def init_carry(func):
    if func == "+":
        return """
    carry := '0';"""
    else:
        return """
    carry := '1';"""


def extract_extend_bit(name, typ):
    res = """
       if i > {0}a'left then
          {0}b := """
    if typ == "UNSIGNED":
        res += "'0';"
    else:
        res += "{0} ({0}'left);"
    res += (
        """
       else
          """
        + extract_bit(name)
        + """
       end if;"""
    )
    return res.format(name)


def disp_vec_vec_binary(func, typ):
    """Generate vector binary function body."""
    res = (
        """
  function "{0}" (l, r : {1}) return {1}
  is
    constant lft : integer := MAX (l'length, r'length) - 1;
    subtype res_type is {1} (lft downto 0);
    alias la : {1} (l'length - 1 downto 0) is l;
    alias ra : {1} (r'length - 1 downto 0) is r;
    variable res : res_type;
    variable lb, rb, carry : """
        + logic_type()
        + """;
  begin
    if la'left < 0 or ra'left < 0 then
       return null_{1};
    end if;"""
    )

    res += init_carry(func)

    res += """
    for i in 0 to lft loop"""
    res += extract_extend_bit("l", typ)
    res += extract_extend_bit("r", typ)

    if logic == "std":
        res += (
            """
       if lb = 'X' or rb = 'X' then"""
            + disp_non_logical_warning(func)
            + """
          res := (others => 'X');
          exit;
        end if;"""
        )
    if func == "-":
        res += """
      rb := not rb;"""
    res += """
      res (i) := compute_sum (carry, rb, lb);
      carry := compute_carry (carry, rb, lb);
    end loop;
    return res;
  end "{0}";
"""
    w(res.format(func, typ))


def declare_int_var(name, typ):
    res = """
     variable {0}1, {0}2 : {1};
     variable {0}d : nat1;"""
    if typ == "INTEGER":
        res += """
     constant {0}msb : nat1 := boolean'pos({0} < 0);"""
    return res.format(name, typ)


def init_int_var(name, typ):
    return """
    {0}1 := {0};""".format(
        name
    )


def extract_int_lsb(name, typ):
    res = """
       {0}2 := {0}1 / 2;"""
    if typ == "INTEGER":
        res += """
       if {0}1 < 0 then
          {0}d := 2 * {0}2 - {0}1;
          {0}1 := {0}2 - {0}d;
        else
          {0}d := {0}1 - 2 * {0}2;
          {0}1 := {0}2;
        end if;"""
    else:
        res += """
       {0}d := {0}1 - 2 * {0}2;
       {0}1 := {0}2;"""
    res += """
       {0}b := nat1_to_01 ({0}d);"""
    return res.format(name, typ)


def check_int_truncated(func, name, typ):
    if typ == "INTEGER":
        v = "-{0}msb".format(name)
    else:
        v = "0"
    return """
    if {1}1 /= {2} then
      assert NO_WARNING
        report "NUMERIC_STD.""{0}"": vector is truncated"
        severity warning;
    end if;""".format(
        func, name, v
    )


def create_vec_int_dict(func, left, right):
    if left in vec_types:
        dic = {"vtype": left, "itype": right, "vparam": "l", "iparam": "r"}
    else:
        dic = {"vtype": right, "itype": left, "vparam": "r", "iparam": "l"}
    dic.update({"ltype": left, "rtype": right, "func": func, "logic": logic_type()})
    return dic


def disp_vec_int_binary(func, left, right):
    """Generate vector binary function body."""
    dic = create_vec_int_dict(func, left, right)
    res = (
        """
  function "{func}" (l : {ltype}; r : {rtype}) return {vtype}
  is
    subtype res_type is {vtype} ({vparam}'length - 1 downto 0);
    alias {vparam}a : res_type is {vparam};"""
        + declare_int_var(dic["iparam"], dic["itype"])
        + """
    variable res : res_type;
    variable lb, rb, carry : {logic};
  begin
    if res'length < 0 then
       return null_{vtype};
    end if;"""
    )

    # Initialize carry.  For subtraction, use 2-complement.
    res += init_carry(func)

    res += (
        init_int_var(dic["iparam"], dic["itype"])
        + """
    for i in res'reverse_range loop
       """
        + extract_bit(dic["vparam"])
        + "\n"
        + extract_int_lsb(dic["iparam"], dic["itype"])
    )

    if logic == "std":
        res += (
            """
       if {vparam}b = 'X' then"""
            + disp_non_logical_warning(func)
            + """
        res := (others => 'X');
        {iparam}1 := 0;
        exit;
      end if;"""
        )

    # 2-complement for subtraction
    if func == "-":
        res += """
      rb := not rb;"""

    res += (
        """
      res (i) := compute_sum (carry, rb, lb);
      carry := compute_carry (carry, rb, lb);
    end loop;"""
        + check_int_truncated(func, dic["iparam"], dic["itype"])
        + """
    return res;
  end "{func}";\n"""
    )
    w(res.format(**dic))


def disp_vec_int_gcompare(func, left, right):
    """Generate comparison function."""
    dic = create_vec_int_dict(func, left, right)
    res = (
        """
  function {func} (l : {ltype}; r : {rtype}) return compare_type
  is
    subtype res_type is {vtype} ({vparam}'length - 1 downto 0);
    alias la : res_type is l;"""
        + declare_int_var(dic["iparam"], dic["itype"])
        + """
    variable lb, rb : {logic};
    variable res : compare_type;
  begin
    res := compare_eq;"""
    )

    res += (
        init_int_var(dic["iparam"], dic["itype"])
        + """
    for i in {vparam}a'reverse_range loop
       """
        + extract_bit(dic["vparam"])
        + extract_int_lsb("r", right)
    )

    if logic == "std":
        res += """
       if {vparam}b = 'X' then
         return compare_unknown;
       end if;"""

    res += """
       if lb = '1' and rb = '0' then
         res := compare_gt;
       elsif lb = '0' and rb = '1' then
         res := compare_lt;
       end if;
    end loop;"""
    if func == "ucompare":
        res += """
    if r1 /= 0 then
       res := compare_lt;
    end if;"""
    else:
        res += (
            """
    if """
            + conv_bit("l (l'left)")
            + """ = '1' then
      if r >= 0 then
         res := compare_lt;
      end if;
    else
      if r < 0 then
         res := compare_gt;
      end if;
    end if;"""
        )
    res += """
    return res;
  end {func};
"""
    w(res.format(**dic))


def disp_vec_int_compare(func, left, right):
    """Generate comparison function."""
    dic = create_vec_int_dict(func, left, right)
    res = (
        """
  function "{func}" (l : {ltype}; r : {rtype}) return boolean
  is
     subtype res_type is {vtype} ({vparam}'length - 1 downto 0);
     alias {vparam}a : res_type is {vparam};"""
        + declare_int_var(dic["iparam"], dic["itype"])
        + """
     variable res : compare_type;
  begin
     if {vparam}'length = 0 then
        assert NO_WARNING
           report "NUMERIC_STD.""{func}"": null argument, returning FALSE"
           severity warning;
        return false;
     end if;

     res := """
    )

    if left == "SIGNED" or right == "SIGNED":
        res += "scompare"
    else:
        res += "ucompare"
    if left in vec_types:
        res += " (l, r);"
    else:
        res += " (r, l);"
    if logic == "std":
        res += (
            """
     if res = compare_unknown then"""
            + disp_non_logical_warning(func)
            + """
      return false;
     end if;"""
        )

    if left in vec_types:
        res += """
     return res {func} compare_eq;"""
    else:
        res += """
     return compare_eq {func} res;"""
    res += """
  end "{func}";
"""
    w(res.format(**dic))


def disp_vec_vec_gcompare(func, typ):
    """Generate comparison function."""
    res = """
  function {func} (l, r : {typ}) return compare_type
  is
     constant sz : integer := MAX (l'length, r'length) - 1;
     alias la : {typ} (l'length - 1 downto 0) is l;
     alias ra : {typ} (r'length - 1 downto 0) is r;
     variable lb, rb : {logic};
     variable res : compare_type;
  begin"""
    if typ == "SIGNED":
        res += (
            """
     --  Consider sign bit as S * -(2**N).
     lb := """
            + conv_bit("la (la'left)")
            + """;
     rb := """
            + conv_bit("ra (ra'left)")
            + """;
     if lb = '1' and rb = '0' then
        return compare_lt;
     elsif lb = '0' and rb = '1' then
        return compare_gt;
     else
        res := compare_eq;
     end if;"""
        )
    else:
        res += """
     res := compare_eq;"""
    if typ == "SIGNED":
        res += """
     for i in 0 to sz - 1 loop"""
    else:
        res += """
    for i in 0 to sz loop"""
    res += extract_extend_bit("l", typ)
    res += extract_extend_bit("r", typ)
    if logic == "std":
        res += """
       if lb = 'X' or rb = 'X' then
         return compare_unknown;
       end if;"""

    res += """
       if lb = '1' and rb = '0' then
         res := compare_gt;
       elsif lb = '0' and rb = '1' then
         res := compare_lt;
       end if;
     end loop;

     return res;
  end {func};\n"""
    w(res.format(func=func, typ=typ, logic=logic_type()))


def disp_vec_vec_compare(func, typ):
    """Generate comparison function."""
    res = """
  function "{func}" (l, r : {typ}) return boolean
  is
     variable res : compare_type;
  begin
     if l'length = 0 or r'length = 0 then
        assert NO_WARNING
           report "NUMERIC_STD.""{func}"": null argument, returning FALSE"
           severity warning;
        return false;
     end if;

     res := """
    if typ == "SIGNED":
        res += "scompare"
    else:
        res += "ucompare"
    res += """ (l, r);"""

    if logic == "std":
        res += (
            """
     if res = compare_unknown then"""
            + disp_non_logical_warning(func)
            + """
       return false;
     end if;"""
        )

    res += """
     return res {func} compare_eq;
  end "{func}";\n"""
    w(res.format(func=func, typ=typ))


def disp_vec_not(typ):
    """Generate vector binary function body."""
    w(
        """
  function "not" (l : {0}) return {0}
  is
    subtype res_type is {0} (l'length - 1 downto 0);
    alias la : res_type is l;
    variable res : res_type;
  begin
    for I in res_type'range loop
      res (I) := not la (I);
    end loop;
    return res;
  end "not";\n""".format(
            typ
        )
    )


def disp_resize(typ):
    res = """
  function resize (ARG : {0}; NEW_SIZE: natural) return {0}
  is
     alias arg1 : {0} (ARG'length - 1 downto 0) is arg;
     variable res : {0} (new_size - 1 downto 0) := (others => '0');
  begin
     if new_size = 0 then
        return null_{0};
     end if;
     if arg1'length = 0 then
        return res;
     end if;
     if arg1'length > new_size then
        --  Reduction."""
    if typ == "SIGNED":
        res += """
        res (res'left) := arg1 (arg1'left);
        res (res'left - 1 downto 0) := arg1 (res'left - 1 downto 0);"""
    else:
        res += """
        res := arg1 (res'range);"""
    res += """
     else
        --  Expansion
        res (arg1'range) := arg1;"""
    if typ == "SIGNED":
        res += """
        res (res'left downto arg1'length) := (others => arg1 (arg1'left));"""
    res += """
    end if;
    return res;
  end resize;\n"""
    w(res.format(typ))


def gen_shift(dir, inv):
    if (dir == "left") ^ inv:
        res = """
       res (res'left downto {opp}count) := arg1 (arg1'left {sub} count downto 0);"""
    else:
        res = """
       res (res'left {sub} count downto 0) := arg1 (arg1'left downto {opp}count);"""
    if inv:
        return res.format(opp="-", sub="+")
    else:
        return res.format(opp="", sub="-")


def disp_shift_op(name, typ, dir):
    res = """
  function {0} (ARG : {1}; COUNT: INTEGER) return {1}
  is
     subtype res_type is {1} (ARG'length - 1 downto 0);
     alias arg1 : res_type is arg;
     variable res : res_type := (others => '0');
  begin
     if res'length = 0 then
        return null_{1};
     end if;
     if count >= 0 and count <= arg1'left then"""
    res += gen_shift(dir, False)
    res += """
     elsif count < 0 and count >= -arg1'left then"""
    res += gen_shift(dir, True)
    res += """
     end if;
     return res;
  end {0};\n"""
    w(res.format(name, typ))


def disp_shift(name, typ, dir):
    res = """
  function {0} (ARG : {1}; COUNT: NATURAL) return {1}
  is
     subtype res_type is {1} (ARG'length - 1 downto 0);
     alias arg1 : res_type is arg;
     variable res : res_type := (others => """
    if typ == "SIGNED" and dir == "right":
        res += "arg1 (arg1'left)"
    else:
        res += "'0'"
    res += """);
  begin
     if res'length = 0 then
        return null_{1};
     end if;
     if count <= arg1'left then"""
    res += gen_shift(dir, False)
    res += """
     end if;
     return res;
  end {0};\n"""
    w(res.format(name, typ))


def disp_rotate(name, typ, dir):
    if "rotate" in name:
        count_type = "natural"
        op = "rem"
    else:
        count_type = "integer"
        op = "mod"
    res = (
        """
  function {0} (ARG : {1}; COUNT: {2}) return {1}
  is
     subtype res_type is {1} (ARG'length - 1 downto 0);
     alias arg1 : res_type is arg;
     variable res : res_type := (others => '0');
     variable cnt : natural;
  begin
     if res'length = 0 then
        return null_{1};
     end if;
     cnt := count """
        + op
        + " res'length;"
    )
    if dir == "left":
        res += """
     res (res'left downto cnt) := arg1 (res'left - cnt downto 0);
     res (cnt - 1 downto 0) := arg1 (res'left downto res'left - cnt + 1);"""
    else:
        res += """
     res (res'left - cnt downto 0) := arg1 (res'left downto cnt);
     res (res'left downto res'left - cnt + 1) := arg1 (cnt - 1 downto 0);"""
    res += """
     return res;
  end {0};\n"""
    w(res.format(name, typ, count_type))


def disp_vec_vec_mul(func, typ):
    res = (
        """
  function "{0}" (L, R : {1}) return {1}
  is
     alias la : {1} (L'Length - 1 downto 0) is l;
     alias ra : {1} (R'Length - 1 downto 0) is r;
     variable res : {1} (L'length + R'Length -1 downto 0) := (others => '0');
     variable rb, lb, vb, carry : """
        + logic_type()
        + """;
  begin
     if la'length = 0 or ra'length = 0 then
        return null_{1};
     end if;
     --  Shift and add L.
     for i in natural range 0 to ra'left """
    )
    if typ == "SIGNED":
        res += "- 1 "
    res += (
        """loop
       """
        + extract_bit("r")
        + """
       if rb = '1' then
          --  Compute res := res + shift_left (l, i).
          carry := '0';
          for j in la'reverse_range loop
            lb := la (j);
            vb := res (i + j);
            res (i + j) := compute_sum (carry, vb, lb);
            carry := compute_carry (carry, vb, lb);
          end loop;"""
    )
    if typ == "UNSIGNED":
        res += """
          --  Propagate carry.
          for j in i + la'length to res'left loop
             exit when carry = '0';
             vb := res (j);
             res (j) := carry xor vb;
             carry := carry and vb;
          end loop;"""
    else:
        res += """
          --  Sign extend and propagate carry.
          lb := la (la'left);
          for j in i + l'length to res'left loop
             vb := res (j);
             res (j) := compute_sum (carry, vb, lb);
             carry := compute_carry (carry, vb, lb);
          end loop;"""
    if logic == "std":
        res += """
       elsif rb = 'X' then""" + disp_non_logical_warning(
            func
        )
    res += """
       end if;
     end loop;"""
    if typ == "SIGNED":
        res += """
     if ra (ra'left) = '1' then
        --  R is a negative number.  It is considered as:
        --   -2**n + (Rn-1 Rn-2 ... R0).
        --  Compute res := res - 2**n * l.
        carry := '1';
        for i in la'reverse_range loop
          vb := res (ra'length - 1 + i);
          lb := not la (i);
          res (ra'length - 1+ i) := compute_sum (carry, vb, lb);
          carry := compute_carry (carry, vb, lb);
        end loop;
        vb := res (res'left);
        lb := not la (la'left);
        res (res'left) := compute_sum (carry, vb, lb);
     end if;"""
    res += """
     return res;
  end "{0}";\n"""
    w(res.format(func, typ))


def disp_vec_int_mul(left, right):
    res = """
  function "*" (L : {0}; R : {1}) return {0}
  is
     constant size : natural := l'length;
  begin
     if size = 0 then
        return null_{0};
     end if;
     return l * to_{0} (r, size);
  end "*";\n"""
    w(res.format(left, right))


def disp_int_vec_mul(left, right):
    res = """
  function "*" (L : {0}; R : {1}) return {1}
  is
     constant size : natural := r'length;
  begin
     if size = 0 then
        return null_{1};
     end if;
     return r * to_{1} (l, size);
  end "*";\n"""
    w(res.format(left, right))


def disp_neg(func):
    res = (
        """
  function "{func}" (ARG : SIGNED) return SIGNED
  is
     subtype arg_type is SIGNED (ARG'length - 1 downto 0);
     alias arga : arg_type is arg;
     variable res : arg_type;
     variable carry, a : """
        + logic_type()
        + """;
  begin
     if arga'length = 0 then
        return null_signed;
     end if;"""
    )
    if logic == "std":
        res += (
            """
     if has_0x (arga) = 'X' then"""
            + disp_non_logical_warning("-")
            + """
       return arg_type'(others => 'X');
     end if;"""
        )
    if func == "abs":
        res += """
     if arga (arga'left) = '0' then
        return arga;
     end if;"""
    res += """
     carry := '1';
     for i in arga'reverse_range loop
       a := not arga (i);
       res (i) := carry xor a;
       carry := carry and a;
     end loop;
     return res;
  end "{func}";\n"""
    w(res.format(func=func))


def disp_has_0x(typ):
    res = """
  function has_0x (a : {0}) return {1}
  is
     variable res : {1} := '0';
  begin
     for i in a'range loop"""
    if logic == "std":
        res += """
      if a (i) = 'X' then
         return 'X';
      end if;"""
    res += """
      res := res or a (i);
    end loop;
    return res;
  end has_0x;\n"""
    w(res.format(typ, logic_type()))


def disp_size():
    w(
        """
  function size_unsigned (n : natural) return natural
  is
     --  At least one bit (even for 0).
     variable res : natural := 1;
     variable n1 : natural := n;
  begin
     while n1 > 1 loop
        res := res + 1;
        n1 := n1 / 2;
     end loop;
     return res;
  end size_unsigned;\n"""
    )

    w(
        """
  function size_signed (n : integer) return natural
  is
     variable res : natural := 1;
     variable n1 : natural;
  begin
     if n >= 0 then
        n1 := n;
     else
        -- Use /N = -X -1 = -(X + 1)  (No overflow).
        n1 := -(n + 1);
     end if;
     while n1 /= 0 loop
        res := res + 1;
        n1 := n1 / 2;
     end loop;
     return res;
  end size_signed;\n"""
    )


def disp_divmod():
    w(
        """
  --  All index range are normalized (N downto 0).
  --  NUM and QUOT have the same range.
  --  DEM and REMAIN have the same range.
  --  No 'X'.
  procedure divmod (num, dem : UNSIGNED; quot, remain : out UNSIGNED)
  is
     --  An extra bit is needed so that it is always possible that DEM >= REG.
     variable reg : unsigned (dem'left + 1 downto 0) := (others => '0');
     variable sub : unsigned (dem'range) := (others => '0');
     variable carry, d : """
        + logic_type()
        + """;
  begin
     for i in num'range loop
        --  Shift to add a new bit from NUM to REG.
        reg (reg'left downto 1) := reg (reg'left - 1 downto 0);
        reg (0) := num (i);
        --  Substract: REG - DEM
        carry := '1';
        for j in dem'reverse_range loop
           d := not dem (j);
           sub (j) := compute_sum (carry, reg (j), d);
           carry := compute_carry (carry, reg (j), d);
        end loop;
        --  Do not forget the extra bit in REG.
        carry := compute_carry (carry, reg (reg'left), '1');
        --  Test
        if carry = '0' then
           --  REG < DEM
           quot (i) := '0';
        else
           --  REG >= DEM: do the substraction
           quot (i) := '1';
           reg (reg'left) := '0';
           reg (sub'range) := sub;
        end if;
     end loop;
     remain := reg (dem'range);
  end divmod;
"""
    )


def disp_vec_vec_udiv(func):
    res = (
        """
  function "{func}" (L, R : UNSIGNED) return UNSIGNED
  is
     subtype l_type is UNSIGNED (L'length - 1 downto 0);
     subtype r_type is UNSIGNED (R'length - 1 downto 0);
     alias la : l_type is l;
     alias ra : r_type is r;
     variable quot : l_type;
     variable rema : r_type;
     variable r0 : """
        + logic_type()
        + """ := has_0x (r);
  begin
     if la'length = 0 or ra'length = 0 then
        return null_unsigned;
     end if;"""
    )
    if logic == "std":
        res += (
            """
     if has_0x (l) = 'X' or r0 = 'X' then"""
            + disp_non_logical_warning("/")
            + """
        return l_type'(others => 'X');
     end if;"""
        )
    res += """
     assert r0 /= '0'
        report "NUMERIC_STD.""{func}"": division by 0"
        severity error;
     divmod (la, ra, quot, rema);"""
    if func == "/":
        res += """
     return quot;"""
    else:
        res += """
     return rema;"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))


def disp_vec_int_udiv(func):
    res = """
  function "{func}" (L : UNSIGNED; R : NATURAL) return UNSIGNED
  is
     constant r_size : natural := size_unsigned (r);
  begin
     if l'length = 0 then
        return null_unsigned;
     end if;"""
    if func in ["mod", "rem"]:
        res += """
     return resize (l {func} to_unsigned (r, r_size), l'length);"""
    else:
        res += """
     return l {func} to_unsigned (r, r_size);"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))

    res = """
  function "{func}" (L : NATURAL; R : UNSIGNED) return UNSIGNED
  is
     constant l_size : natural := size_unsigned (l);
  begin
     if r'length = 0 then
        return null_unsigned;
     end if;"""
    if func == "/":
        res += """
     return resize (to_unsigned (l, l_size) {func} r, r'length);"""
    else:
        res += """
     return to_unsigned (l, l_size) {func} r;"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))


def disp_vec_vec_sdiv(func):
    res = (
        """
  function "{func}" (L, R : SIGNED) return SIGNED
  is
     subtype l_type is SIGNED (L'length - 1 downto 0);
     subtype r_type is SIGNED (R'length - 1 downto 0);
     alias la : l_type is l;
     alias ra : r_type is r;
     subtype l_utype is UNSIGNED (l_type'range);
     subtype r_utype is UNSIGNED (r_type'range);
     variable lu : l_utype;
     variable ru : r_utype;
     variable quot : l_utype;
     variable rema : r_utype;
     variable r0 : """
        + logic_type()
        + """ := has_0x (r);
  begin
     if la'length = 0 or ra'length = 0 then
        return null_signed;
     end if;"""
    )
    if logic == "std":
        res += (
            """
     if has_0x (l) = 'X' or r0 = 'X' then"""
            + disp_non_logical_warning(func)
            + """
        return l_type'(others => 'X');
     end if;"""
        )
    res += """
     assert r0 /= '0'
        report "NUMERIC_STD.""{func}"": division by 0"
        severity error;"""
    res += """
     if la (la'left) = '1' then
        lu := unsigned (-la);
     else
        lu := unsigned (la);
     end if;
     if ra (ra'left) = '1' then
        ru := unsigned (-ra);
     else
        ru := unsigned (ra);
     end if;
     divmod (lu, ru, quot, rema);"""
    if func == "/":
        res += """
     if (ra (ra'left) xor la (la'left)) = '1' then
       return -signed (quot);
     else
       return signed (quot);
     end if;"""
    elif func == "rem":
        res += """
     --  Result of rem has the sign of the dividend.
     if la (la'left) = '1' then
       return -signed (rema);
     else
       return signed (rema);
     end if;"""
    elif func == "mod":
        res += """
     --  Result of mod has the sign of the divisor.
     if rema = r_utype'(others => '0') then
        --  If the remainder is 0, then the modulus is 0.
        return signed (rema);
     else
        if ra (ra'left) = '1' then
           if la (la'left) = '1' then
              return -signed (rema);
           else
              return ra + signed (rema);
           end if;
        else
           if la (la'left) = '1' then
              return ra - signed (rema);
           else
              return signed (rema);
           end if;
        end if;
     end if;"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))


def disp_vec_int_sdiv(func):
    res = """
  function "{func}" (L : SIGNED; R : INTEGER) return SIGNED
  is
     constant r_size : natural := size_signed (r);
  begin
     if l'length = 0 then
        return null_signed;
     end if;"""
    if func == "/":
        res += """
     return l {func} to_signed (r, r_size);"""
    else:
        res += """
     return resize (l {func} to_signed (r, r_size), l'length);"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))

    res = """
  function "{func}" (L : INTEGER; R : SIGNED) return SIGNED
  is
     constant l_size : natural := size_signed (l);
  begin
     if r'length = 0 then
        return null_signed;
     end if;"""
    if func == "/":
        res += """
     return resize (to_signed (l, max (l_size, r'length)) {func} r, r'length);"""
    else:
        res += """
     return to_signed (l, l_size) {func} r;"""
    res += """
  end "{func}";\n"""
    w(res.format(func=func))


def disp_all_log_funcs():
    """Generate all function bodies for logic operators."""
    for t in vec_types:
        disp_resize(t)
    for v in vec_types:
        disp_vec_not(v)
    for f in binary_funcs:
        for v in vec_types:
            disp_vec_binary(f, v)
    disp_vec_vec_gcompare("ucompare", "UNSIGNED")
    disp_vec_vec_gcompare("scompare", "SIGNED")
    disp_vec_int_gcompare("ucompare", "UNSIGNED", "NATURAL")
    disp_vec_int_gcompare("scompare", "SIGNED", "INTEGER")
    for f in compare_funcs:
        disp_vec_vec_compare(f, "UNSIGNED")
        disp_vec_vec_compare(f, "SIGNED")
        disp_vec_int_compare(f, "UNSIGNED", "NATURAL")
        disp_vec_int_compare(f, "NATURAL", "UNSIGNED")
        disp_vec_int_compare(f, "SIGNED", "INTEGER")
        disp_vec_int_compare(f, "INTEGER", "SIGNED")
    for t in vec_types:
        for d in ["left", "right"]:
            disp_shift("shift_" + d, t, d)
        for d in ["left", "right"]:
            disp_rotate("rotate_" + d, t, d)
        if v93:
            disp_shift_op('"sll"', t, "left")
            disp_shift_op('"srl"', t, "right")
            disp_rotate('"rol"', t, "left")
            disp_rotate('"ror"', t, "right")


def disp_match(typ):
    res = """
  function std_match (l, r : {0}) return boolean
  is
     alias la : {0} (l'length downto 1) is l;
     alias ra : {0} (r'length downto 1) is r;
  begin
     if la'left = 0 or ra'left = 0 then
        assert NO_WARNING
          report "NUMERIC_STD.STD_MATCH: null argument, returning false"
          severity warning;
        return false;
     elsif la'left /= ra'left then
        assert NO_WARNING
          report "NUMERIC_STD.STD_MATCH: args length mismatch, returning false"
          severity warning;
        return false;
     else
        for i in la'range loop
           if not match_table (la (i), ra (i)) then
             return false;
           end if;
        end loop;
        return true;
     end if;
  end std_match;\n"""
    w(res.format(typ))


def disp_all_match_funcs():
    disp_match("std_ulogic_vector")
    disp_match("std_logic_vector")
    disp_match("UNSIGNED")
    disp_match("SIGNED")


def disp_all_arith_funcs():
    """Generate all function bodies for logic operators."""
    for op in ["+", "-"]:
        disp_vec_vec_binary(op, "UNSIGNED")
        disp_vec_vec_binary(op, "SIGNED")
        disp_vec_int_binary(op, "UNSIGNED", "NATURAL")
        disp_vec_int_binary(op, "NATURAL", "UNSIGNED")
        disp_vec_int_binary(op, "SIGNED", "INTEGER")
        disp_vec_int_binary(op, "INTEGER", "SIGNED")
    disp_vec_vec_mul("*", "UNSIGNED")
    disp_vec_vec_mul("*", "SIGNED")
    disp_vec_int_mul("UNSIGNED", "NATURAL")
    disp_vec_int_mul("SIGNED", "INTEGER")
    disp_int_vec_mul("NATURAL", "UNSIGNED")
    disp_int_vec_mul("INTEGER", "SIGNED")
    disp_has_0x("UNSIGNED")
    disp_divmod()
    disp_size()
    disp_vec_vec_udiv("/")
    disp_vec_int_udiv("/")
    disp_vec_vec_udiv("rem")
    disp_vec_int_udiv("rem")
    disp_vec_vec_udiv("mod")
    disp_vec_int_udiv("mod")
    disp_has_0x("SIGNED")
    disp_neg("-")
    disp_neg("abs")
    disp_vec_vec_sdiv("/")
    disp_vec_int_sdiv("/")
    disp_vec_vec_sdiv("rem")
    disp_vec_int_sdiv("rem")
    disp_vec_vec_sdiv("mod")
    disp_vec_int_sdiv("mod")


# Patterns to replace
pats = {
    "  @LOG\n": disp_all_log_funcs,
    "  @ARITH\n": disp_all_arith_funcs,
    "  @MATCH\n": disp_all_match_funcs,
}

spec_file = "numeric_std.vhdl"
# proto_file='numeric_std-body.proto'


def gen_body(proto_file):
    w("--  This -*- vhdl -*- file was generated from " + proto_file + "\n")
    for line in open(proto_file):
        if line in pats:
            pats[line]()
            continue
        w(line)


# Copy spec
for log in logics:
    for std in ["87", "93"]:
        out = open("v" + std + "/numeric_" + log + ".vhdl", "w")
        for line in open("numeric_" + log + ".proto"):
            if line == "  @COMMON\n":
                for lcom in open("numeric_common.proto"):
                    if lcom[0:2] == "--":
                        pass
                    elif std == "87" and (
                        '"xnor"' in lcom
                        or '"sll"' in lcom
                        or '"srl"' in lcom
                        or '"rol"' in lcom
                        or '"ror"' in lcom
                    ):
                        w("--" + lcom[2:])
                    else:
                        w(lcom)
            else:
                w(line)
        out.close()

# Generate bodies
v93 = False
for l in logics:
    logic = l
    out = open("v87/numeric_{0}-body.vhdl".format(l), "w")
    gen_body("numeric_{0}-body.proto".format(l))
    out.close()

v93 = True
binary_funcs.append("xnor")
for l in logics:
    logic = l
    out = open("v93/numeric_{0}-body.vhdl".format(l), "w")
    gen_body("numeric_{0}-body.proto".format(l))
    out.close()
